# Import relevant libraries and dependencies
from typing import List, Union, Dict, Tuple
import numpy as np
from .wasserstein import WassersteinDistance

"""
add string2string module code, src = https://github.com/stanfordnlp/string2string

MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# Parent class for all the string algorithms implemented in this module
class StringAlgs:
    """
        This class is the parent class for all the string algorithms implemented in this module.
    """
    # Initialize the class
    def __init__(self,
                match_weight: float = 0.0,
        ) -> None:
        # Set the match weight
        self.match_weight = match_weight

# Levenshtein edit distance class
class LevenshteinEditDistance(StringAlgs):
    def __init__(self,
                match_weight: float = 0.0,
                insert_weight: float = 1.0,
                delete_weight: float = 1.0,
                substitute_weight: float = 1.0,
    ) -> None:
        r"""
        This class initializes the Levenshtein edit distance algorithm. Levenshtein edit distance represents the minimum number of edit distance operations (insertion, deletion, and substitution) required to convert one string to another.
            
        The Levenshtein edit distance (with unit cost for each edit distance operation) is given by the following recurrence relation: 

        .. math::
            :nowrap:

            \begin{align}
            d[i, j] := \min( & d[i-1, j-1] + \texttt{mismatch}(i, j),  \\
                                & d[i-1, j] + 1,  \\
                                & d[i, j-1] + 1),
            \end{align}

        where :math:`\texttt{mismatch}(i, j)` is 1 if the i-th element in str1 is not equal to the j-th element in str2, and 0 otherwise.

        Arguments:
            match_weight (float): The weight of a match (default: 0.0).
            insert_weight (float): The weight of an insertion (default: 1.0).
            delete_weight (float): The weight of a deletion (default: 1.0).
            substitute_weight (float): The weight of a substitution (default: 1.0).

        Raises:
            AssertionError: If any of the weights are negative.
        """
        # Set the match weight
        super().__init__(match_weight=match_weight)

        # Set the insert, delete, and substite weights
        self.insert_weight = insert_weight
        self.delete_weight = delete_weight
        self.substitute_weight = substitute_weight

        # Assert that all the weights are non-negative
        assert min(match_weight, insert_weight, delete_weight, substitute_weight) >= 0.0


    
    # Compute the Levenshtein edit distance between two strings using recursion
    def compute_recursive(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
    ) -> float:
        r"""
        This function computes the Levenshtein edit distance between two strings (or lists of strings) using recursion.

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).

        Returns:
            The Levenshtein edit distance between the two strings.

        .. note::
            * The solution presented here utilizes recursion to compute the Levenshtein edit distance between two strings. It has an exponential time complexity and is not recommended for pairs of strings with a large length.
            * The time complexity of this function is :math:`O(3^{m+n})`, where :math:`m` and :math:`n` are the lengths of the two strings.
        """
        # Base case
        if len(str1) == 0:
            return len(str2) * self.insert_weight
        elif len(str2) == 0:
            return len(str1) * self.delete_weight

        # Compute the mismatch
        mismatch = 0.0 if str1[-1] == str2[-1] else self.substitute_weight

        # Compute the Levenshtein edit distance
        return min(
            self.compute_recursive(str1[:-1], str2[:-1]) + mismatch,
            self.compute_recursive(str1[:-1], str2) + self.delete_weight,
            self.compute_recursive(str1, str2[:-1]) + self.insert_weight,
        )

    

    # Compute the Levenshtein edit distance between two strings using memoization
    def compute_recursive_memoization(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
    ) -> float:
        r"""
        This function computes the Levenshtein edit distance between two strings (or lists of strings) using memoization.

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).

        Returns:
            The Levenshtein edit distance between the two strings.

        .. note::
            * The solution presented here utilizes memoization to compute the Levenshtein edit distance between two strings. 
            * The time complexity of this function is :math:`\mathcal{O}(m n)`, where :math:`m` and :math:`n` are the lengths of the two strings.
        """
        # Initialize the memoization dictionary
        memoization = {}

        # Compute the Levenshtein edit distance
        return self.compute_memoization_helper(str1, str2, memoization)

    

    # Compute the Levenshtein edit distance between two strings using memoization (helper function)
    def compute_memoization_helper(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
        memoization: Dict[Tuple[str, str], float],
    ) -> float:
        r"""
        This is a helper function that computes the Levenshtein edit distance between two strings (or lists of strings) using memoization.

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).
            memoization (dict): The memoization dictionary.

        Returns:
            The Levenshtein edit distance between the two strings.

        .. note::
            * The solution presented here utilizes memoization to compute the Levenshtein edit distance between two strings.
            * One can also use the :func:`functools.lru_cache` (@lru_cache()) decorator to memoize the function calls. However, for the sake of educational purposes, we have implemented memoization using a dictionary.
            * The time complexity of this function is quadratic, that is :math:`\mathcal{O}(nm)`, where m and n are the lengths of the two strings.
        """
        # Base case
        if len(str1) == 0:
            return len(str2) * self.insert_weight
        elif len(str2) == 0:
            return len(str1) * self.delete_weight

        # Check if the Levenshtein edit distance has already been computed
        if (str1, str2) in memoization:
            return memoization[(str1, str2)]

        # Compute the mismatch
        mismatch = 0.0 if str1[-1] == str2[-1] else self.substitute_weight

        # Compute the Levenshtein edit distance
        memoization[(str1, str2)] = min(
            self.compute_memoization_helper(str1[:-1], str2[:-1], memoization) + mismatch,
            self.compute_memoization_helper(str1[:-1], str2, memoization) + self.delete_weight,
            self.compute_memoization_helper(str1, str2[:-1], memoization) + self.insert_weight,
        )

        # Return the Levenshtein edit distance
        return memoization[(str1, str2)]



    # Compute the Levenshtein edit distance between two strings using dynamic programming
    def compute_dynamic_programming(self,
        str1: Union[str, List[str]], 
        str2: Union[str, List[str]],
    ) -> float:
        r"""
        This function computes the Levenshtein edit distance between two strings (or lists of strings) using dynamic programming (Wagner-Fischer algorithm).

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).

        Returns:
            The Levenshtein edit distance between the two strings.

        .. note::
            * The solution presented here utilizes dynamic programming principles to compute the Levenshtein edit distance between two strings. 
            * This solution is also known as the Wagner-Fischer algorithm. [WF1974]_
            * The time complexity of this dynamic-programming-based solution is :math:`\mathcal{O}(nm)`, and the space complexity is :math:`\mathcal{O}(nm)`, where n and m are the lengths of the two strings, respectively.
            * However, by using only two rows of the distance matrix at a time, the space complexity of the dynamic programming solution can be reduced to :math:`\mathcal{O}(min(n, m))`.
            * The time complexity cannot be made strongly subquadratic time unless SETH is false. [BI2015]_
            * Finally, we note that this solution can be extended to cases where each edit distance operation has a non-unit cost.

            .. [WF1974] Wagner, R.A. and Fischer, M.J., 1974. The string-to-string correction problem. Journal of the ACM (JACM), 21(1), pp.168-173.
            .. [BI2015] Backurs, A. and Indyk, P., 2015, June. Edit distance cannot be computed in strongly subquadratic time (unless SETH is false). In Proceedings of the forty-seventh annual ACM symposium on Theory of computing (pp. 51-58).
        """
        # Lengths of strings str1 and str2, respectively.
        n = len(str1)
        m = len(str2)

        # Initialize the distance matrix.
        dist = np.zeros((n + 1, m + 1))
        for i in range(1, n + 1):
            dist[i, 0] = self.delete_weight * i
        for j in range(1, m + 1):
            dist[0, j] = self.insert_weight * j

        # Dynamic programming step, where each operation has a unit cost:
        # d[i, j] := min(d[i-1, j-1] + mismatch(i, j), d[i-1, j] + 1, d[i, j-1] + 1),
        # where mismatch(i, j) is 1 if str1[i] != str2[j] and 0 otherwise.
        for i in range(1, n + 1):
            for j in range(1, m + 1):
                # Compute the minimum edit distance between str1[:i] and str2[:j].
                dist[i, j] = min(
                    dist[i-1, j-1] + (self.substitute_weight if str1[i-1] != str2[j-1] else self.match_weight),
                    dist[i-1, j] + self.delete_weight, 
                    dist[i, j-1] + self.insert_weight,
                )

        # Return the Levenshtein edit distance between str1 and str2.
        return dist[n, m]



    # Compute the Levenshtein edit distance between two strings
    def compute(self,
        str1: Union[str, List[str]], 
        str2: Union[str, List[str]],
        method: str = "dynamic-programming",
    ) -> float:
        r"""
        This function computes the Levenshtein edit distance between two strings (or lists of strings), using the method specified by the user. 

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).
            method (str): The method to use to compute the Levenshtein edit distance (default: "dynamic-programming").

        Returns:
            The Levenshtein edit distance between the two strings.

        .. note::
            * The method can be one of the following:
                * "recursive": This method computes the Levenshtein edit distance using recursion.
                * "recursive-memoization": This method computes the Levenshtein edit distance using recursion with memoization.
                * "dynamic-programming": This method computes the Levenshtein edit distance using dynamic programming (Wagner-Fischer algorithm).
            * By default, the method is "dynamic-programming".
            
        """
        # If the method is dynamic programming, then compute the Levenshtein edit distance using dynamic programming
        if method == "recursive":
            return self.compute_recursive(str1, str2)
        elif method == "recursive-memoization":
            return self.compute_recursive_memoization(str1, str2)
        return self.compute_dynamic_programming(str1, str2)

# Hamming (edit) distance class
class HammingDistance(StringAlgs):
    def __init__(self, 
                match_weight: float = 0.0,
                substitute_weight: float = 1.0,
        ) -> None:
        r"""
        This function initializes the class variables of the Hamming distance. 
        
        The Hamming distance is the number of positions at which the corresponding symbols are different. [H1950]_

        Arguments:
            match_weight (float): The weight of a match (default: 0.0).
            substitute_weight (float): The weight of a substitution (default: 1.0).

        Raises:
            AssertionError: If the substite weight is negative.

        .. note::
            * The Hamming distance has a time complexity of :math:`\mathcal{O}(n)`, where :math: `n` the length of the two strings.

        .. [H1950] Hamming, R.W., 1968. Error detecting and error correcting codes. Bell System Technical Journal, 29(2), pp.147-160.
        """
        # Set the match weight
        super().__init__(match_weight=match_weight)

        # Set the substite weight
        self.substitute_weight = substitute_weight

        # Assert that the substite weight is non-negative
        assert substitute_weight >= 0.0



    # Compute the Hamming distance between two strings
    def compute(self,
        str1: Union[str, List[str]], 
        str2: Union[str, List[str]],
    ) -> float:
        """
        This function computes the Hamming distance between two strings (or lists of strings).

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).

        Returns:
            The Hamming distance between the two strings.

        Raises:
            ValueError: If the two strings (or lists of strings) have different lengths.
        """

        # Lengths of strings str1 and str2, respectively.
        n = len(str1)
        m = len(str2)

        # Assert that the two strings have the same length
        if n != m:
            raise ValueError("The two strings (or lists of strings) must have the same length.")

        # Compute the Hamming edit distance between str1 and str2.
        return sum(
            self.substitute_weight if str1[i] != str2[i] else self.match_weight
            for i in range(n)
        )


# Damerau-Levenshtein edit distance class
class DamerauLevenshteinDistance(LevenshteinEditDistance):
    def __init__(self, 
                match_weight: float = 0.0,
                insert_weight: float = 1.0,
                delete_weight: float = 1.0,
                substitute_weight: float = 1.0,
                adjacent_transpose_weight: float = 1.0,
        ) -> None:
        r"""
        This function initializes the class variables of the Damerau-Levenshtein distance.
         
        The Damerau-Levenshtein distance is the minimum number of insertions, deletions, substitutions, and transpositions required to transform one string into the other. [D1964]_

        Arguments:
            match_weight (float): The weight of a match (default: 0.0).
            insert_weight (float): The weight of an insertion (default: 1.0).
            delete_weight (float): The weight of a deletion (default: 1.0).
            substitute_weight (float): The weight of a substitution (default: 1.0).
            adjacent_transpose_weight (float): The weight of an adjacent transposition (default: 1.0).

        Raises:
            AssertionError: If the insert, delete, substite, or adjacent transpose weights are negative.

        .. [D1964] Damerau, F.J., 1964. A technique for computer detection and correction of spelling errors. Communications of the ACM, 7(3), pp.171-176.
        """
        # Set the weights of the distance operations
        super().__init__(
            match_weight=match_weight,
            insert_weight=insert_weight,
            delete_weight=delete_weight,
            substitute_weight=substitute_weight,
        )

        # Set the adjacent transpose weight
        self.adjacent_transpose_weight = adjacent_transpose_weight

        # Assert that the adjacent transpose weight is non-negative
        assert adjacent_transpose_weight >= 0.0



    # Compute the Damerau-Levenshtein edit distance between two strings
    def compute(self,
        str1: Union[str, List[str]], 
        str2: Union[str, List[str]],
    ) -> float:
        """
        This function computes the Damerau-Levenshtein edit distance between two strings (or lists of strings).

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).

        Returns:
            The Damerau-Levenshtein distance between the two strings.

        .. note::
            * The Damerau-Levenshtein distance is a variant of the Levenshtein distance that allows for adjacent transpositions.
            * The dynamic programming solution to the Damerau-Levenshtein distance has a time complexity of :math:`\mathcal{O}(nm)`, where n and m are the lengths of the two strings.
        """

        # Lengths of strings str1 and str2, respectively.
        n = len(str1)
        m = len(str2)

        # Initialize the distance matrix.
        dist = np.zeros((n + 1, m + 1))
        for i in range(1, n + 1):
            dist[i, 0] = self.delete_weight * i
        for j in range(1, m + 1):
            dist[0, j] = self.insert_weight * j

        # Dynamic programming solution to the Damerau-Levenshtein edit distance is very similar to that of the Levenshtein edit distance.
        for i in range(1, n + 1):
            for j in range(1, m + 1):
                dist[i, j] = min(
                    dist[i-1, j-1] + (self.substitute_weight if str1[i-1] != str2[j-1] else self.match_weight),
                    dist[i-1, j] + self.delete_weight, 
                    dist[i, j-1] + self.insert_weight,
                )
                # This is the only difference between the Damerau-Levenshtein edit distance and the Levenshtein edit distance.
                if i > 1 and j > 1 and str1[i-1] == str2[j-2] and str1[i-2] == str2[j-1]:
                    dist[i, j] = min(dist[i, j], dist[i-2, j-2] + self.adjacent_transpose_weight)

        # Return the Damerau-Levenshtein edit distance between str1 and str2.
        return dist[n, m]



# -*- coding: utf-8 -*-
#cython: embedsignature=True
#cython: profile=True
#cython: language_level=3

"""
A neural network for NLP tagging tasks.
It employs feature tables to store feature vectors for each token.
"""

import numpy as np
cimport numpy as np
cimport cython
from cpython cimport bool

#from itertools import izip
import logging

#ctypedef float FLOAT_t
ctypedef np.float_t FLOAT_t
#ctypedef int INT_t
ctypedef np.int_t INT_t
ctypedef np.double_t DOUBLE_t

# ----------------------------------------------------------------------
# Math functions

cdef logsumexp(np.ndarray a, axis=None):
    """Compute the log of the sum of exponentials of input elements.
    like: scipy.misc.logsumexp

    Parameters
    ----------
    a : array_like
        Input array.
    axis : int, optional
        Axis over which the sum is taken. By default `axis` is None,
        and all elements are summed.

    Returns
    -------
    res : ndarray
        The result, ``np.log(np.sum(np.exp(a)))`` calculated in a numerically
        more stable way.
    """
    if axis is None:
        a = a.ravel()
    else:
        a = np.rollaxis(a, axis)
    a_max = a.max(axis=0)
    return np.log(np.sum(np.exp(a - a_max), axis=0)) + a_max

cdef hardtanh(np.ndarray weights, inplace=False):
    """
    Hard hyperbolic tangent.
    If inplace is True, modifies the input weights, which will be faster.
    """
    if inplace:
        out = weights
    else:
        out = np.copy(weights)
    inds_greater = weights > 1
    inds_lesser = weights < -1
    out[inds_greater] = 1
    out[inds_lesser] = -1
    
    return out

cdef hardtanhd(np.ndarray[FLOAT_t, ndim=2] weights):
    """derivative of hardtanh"""
    cdef np.ndarray out = np.zeros_like(weights)
    inds = np.logical_and(-1.0 <= weights, weights <= 1.0)
    out[inds] = 1.0
    
    return out

# ----------------------------------------------------------------------

cdef class Network:
    
    # sizes and learning rates
    cdef readonly int word_window_size, input_size, hidden_size, output_size
    cdef public float learning_rate, learning_rate_features
    cdef public float decay_factor
    cdef public bool use_learning_rate_decay
    cdef readonly int features_per_token
    
    # lookup for fast access to all the token embeddings in a sentence
    cdef np.ndarray sentence_lookup
    
    # padding stuff
    cdef np.ndarray padding_left, padding_right
    cdef public np.ndarray pre_padding, pos_padding
    
    # weights, biases, calculated values
    cdef readonly np.ndarray hidden_weights, output_weights
    cdef readonly np.ndarray hidden_bias, output_bias
    cdef readonly np.ndarray input_values, hidden_values, layer2_values
    
    # feature tables
    cdef public list feature_tables
    
    # transitions
    cdef public float learning_rate_trans
    cdef public np.ndarray transitions
    
    # the score for a given path
    cdef readonly float answer_score
    
    # gradients
    cdef readonly np.ndarray net_gradients, trans_gradients
    cdef readonly np.ndarray input_sent_values, hidden_sent_values, layer2_sent_values
    
    # data for statistics during training. 
    cdef float error, accuracy, float_errors, sentence_accuracy
    cdef int num_tokens, skips
        
    # file where the network is saved
    cdef public str network_filename
    cdef public str network_text_filename
    
    # validation
    cdef list validation_sentences
    cdef list validation_tags

    @classmethod
    def create_new(cls, feature_tables, int word_window, int hidden_size, 
                 int output_size):
        """
        Creates a new neural network.
        """
        # sum the number of features in all tables 
        cdef int input_size = sum(table.shape[1] for table in feature_tables)
        input_size *= word_window
        
        # creates the weight matrices

        # set the seed for replicability
        #np.random.seed(42)

        # SENNA: centered uniform distribution with variance = 1/sqrt(fanin)
        # variance = 1/12 interval ^ 2
        # interval = 3.46 / fanin ^ 1/4
        #high = 1.732 / np.power(input_size, 0.25) # SENNA: 0.416
        high = 2.38 / np.sqrt(input_size) # [Bottou-88]
        #high = 0.1              # Fonseca
        hidden_weights = np.random.uniform(-high, high, (hidden_size, input_size))
        hidden_bias = np.random.uniform(-high, high, (hidden_size))
        #high = 1.732 / np.power(hidden_size, 0.25) # SENNA
        high = 2.38 / np.sqrt(hidden_size) # [Bottou-88]
        #high = 0.1              # Fonseca
        output_weights = np.random.uniform(-high, high, (output_size, hidden_size))
        output_bias = np.random.uniform(-high, high, (output_size))
        
        high = 1.0
        # +1 is due for the initial transition
        transitions = np.random.uniform(-high, high, (output_size + 1, output_size))

        net = Network(word_window, input_size, hidden_size, output_size,
                      hidden_weights, hidden_bias, output_weights, output_bias,
                      transitions)
        net.feature_tables = feature_tables
        
        return net
        
    def __init__(self, word_window, input_size, hidden_size, output_size,
                 hidden_weights, hidden_bias, output_weights, output_bias,
                 transitions=None):
        """
        This function isn't expected to be directly called.
        Instead, use the classmethods load_from_file or 
        create_new.
        
        :param transitions: transition weights. If None uses
            Window Level Likelihood instead of Sentence Level Likelihood.
        """
        self.learning_rate = 0
        self.learning_rate_features = 0
        
        self.word_window_size = word_window
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.features_per_token = input_size / word_window
        
        # A_i_j score for jumping from tag i to j
        # A_0_i = transitions[-1]
        self.transitions = transitions

        self.hidden_weights = hidden_weights
        self.hidden_bias = hidden_bias
        self.output_weights = output_weights
        self.output_bias = output_bias
        
        self.validation_sentences = None
        self.validation_tags = None
        
        self.use_learning_rate_decay = False

    def description(self):
        """
        Returns a textual description of the network.
        """
        table_dims = [str(t.shape[1]) for t in self.feature_tables]
        table_dims =  ', '.join(table_dims)
        
        desc = """
Word window size: %d
Feature table sizes: %s
Input layer size: %d
Hidden layer size: %d
Output size: %d
""" % (self.word_window_size, table_dims, self.input_size, self.hidden_size, self.output_size)
        
        return desc
    
    def _create_sentence_lookup(self, np.ndarray sentence):
        """
        Create a lookup matrix with the embeddings values for all tokens in a sentence.
        """        
        #print (sentence)
        cdef np.ndarray padded_sentence = np.concatenate((self.pre_padding,
                                                          sentence,
                                                          self.pos_padding))
        
        # make sure it works on 32 bit python installations
        padded_sentence = padded_sentence.astype(int)
        #padded_sentence = padded_sentence.astype(np.int64)
        
        self.sentence_lookup = np.empty((len(padded_sentence), self.features_per_token))
        ind_from = 0
        
        for i, table in enumerate(self.feature_tables):
            num_dims = table.shape[1]
            ind_to = ind_from + num_dims
            
            token_indices = padded_sentence[:, i]
            embeddings = table.take(token_indices, axis=0)
            self.sentence_lookup[:, ind_from:ind_to] = embeddings
            
            ind_from = ind_to
    
    property padding_left:
        """
        The padding element filling the "void" before the beginning
        of the sentence.
        """
        def __get__(self):
            return self.padding_left
    
        def __set__(self, np.ndarray padding_left):
            self.padding_left = padding_left
            self.pre_padding = np.array((self.word_window_size // 2) * [padding_left])
    
    property padding_right:
        """
        The padding element filling the "void" after the end
        of the sentence.
        """
        def __get__(self):
            return self.padding_right
    
        def __set__(self, np.ndarray padding_right):
            self.padding_right = padding_right
            self.pos_padding = np.array((self.word_window_size // 2) * [padding_right])
    
    def tag_sentence(self, np.ndarray sentence):
        """
        Runs the network for each element in the sentence and returns 
        the sequence of tags.
        
        :param sentence: a 2-dim numpy array, where each item encodes a token.
        """
        scores = self._tag_sentence(sentence)
        # computes full score, combining ftheta and A (if SLL)
        return self._viterbi(scores)

    def _tag_sentence(self, np.ndarray sentence, tags=None):
        """
        Runs the network for each element in the sentence and returns 
        the sequence of tags.
        
        :param sentence: a 2-dim numpy array, where each item encodes a token.
        :param tags: the correct tags (needed when training)
        :return: a (len(sentence), output_size) array with the scores for all tokens
        """
        cdef np.ndarray answer
        cdef np.ndarray input_data
        # scores[t, i] = ftheta_i,t = score for i-th tag, t-th word
        cdef np.ndarray scores = np.empty((len(sentence), self.output_size))
        
        training = tags is not None
        if training:
            self.input_sent_values = np.empty((len(sentence), self.input_size))
            # layer2_values at each token in the correct path
            self.layer2_sent_values = np.empty((len(sentence), self.hidden_size))
            # hidden_values at each token in the correct path
            self.hidden_sent_values = np.empty((len(sentence), self.hidden_size))
        
        self._create_sentence_lookup(sentence)

        # run through all windows in the sentence
        for i in range(len(sentence)):
            input_data = self.sentence_lookup[i:i + self.word_window_size].flatten()
            
            # (hidden_size, input_size) . input_size = hidden_size
            self.layer2_values = self.hidden_weights.dot(input_data) + self.hidden_bias
            self.hidden_values = hardtanh(self.layer2_values, inplace=not training)
            output = self.output_weights.dot(self.hidden_values) + self.output_bias
            scores[i] = output
            
            if training:
                self.input_sent_values[i] = input_data
                self.layer2_sent_values[i] = self.layer2_values
                self.hidden_sent_values[i] = self.hidden_values
        
        if training:
            if self._calculate_gradients_sll(tags, scores):
                self._backpropagate(sentence)

        return scores
    
    def _calculate_delta(self, scores):
        """
        Calculates a matrix with the scores for all possible paths at all given
        points (tokens).
        In the returned matrix, delta[i][j] means the sum of all scores 
        ending in token i with tag j (delta_i(j) in eq. 14 in the paper)
        """
        # logadd for first token. the transition score of the starting tag must be used.
        # it turns out that logadd = log(exp(score)) = score
        # (use long double because taking exp's leads to very very big numbers)
        # scores[t][k] = ftheta_k,t
        delta = scores
        
        # transitions[-1] represents initial transition, A_0,i in paper (mispelled as A_i,0)
        # delta_0(k) = ftheta_k,0 + A_0,i
        delta[0] += self.transitions[-1]
        
        # logadd for the remaining tokens
        # delta_t(k) = ftheta_k,t + logadd_i(delta_t-1(i) + A_i,k)
        #            = ftheta_k,t + log(Sum_i(exp(delta_t-1(i) + A_i,k)))
        transitions = self.transitions[:-1].T # A_k,i
        
        for token in xrange(1, len(delta)):
            # sum by rows
            logadd = logsumexp(delta[token - 1] + transitions, 1)
            delta[token] += logadd
        
        return delta

    @cython.boundscheck(False)
    def _calculate_gradients_sll(self, tags, scores):
        """
        Calculates the output and transition deltas for each token, using Sentence Level Likelihood.
        The aim is to minimize the cost:
        C(theta,A) = logadd(scores for all possible paths) - score(correct path)
        
        :returns: if True, normal gradient calculation was performed.
            If False, the error was too low and weight correction should be
            skipped.
        """
        cdef np.ndarray[DOUBLE_t, ndim=2] delta # (len(sentence), output_size)
        cdef np.ndarray[DOUBLE_t, ndim=2] delta_softmax # (output_size, output_size)
        
        # ftheta_i,t = network output for i-th tag, at t-th word
        # s = Sum_i(A_tags[i-1],tags[i] + ftheta_i,i), i < len(sentence)   (12)
        correct_path_score = 0
        last_tag = self.output_size
        for tag, net_scores in zip(tags[:-2], scores[:-2]):
            #print(tags, last_tag, tag, self.transitions)
            trans = 0 if self.transitions is None else self.transitions[last_tag, tag]
            #print ('trans:',trans)
            #print ('net:',tag, net_scores[tag])
            #print
            correct_path_score += trans + net_scores[tag]
            last_tag = tag 
        
        # delta[t] = delta_t in equation (14)
        delta = self._calculate_delta(scores)
        # logadd_i(delta_T(i)) = log(Sum_i(exp(delta_T(i))))
        # Sentence-level Log-Likelihood (SLL)
        # C(ftheta,A) = logadd_j(s(x, j, theta, A)) - score(correct path)
        #print ': ', np.log(np.sum(np.exp(delta[-1])))
        #print '> ', correct_path_score
        error = np.log(np.sum(np.exp(delta[-1]))) - correct_path_score
        #print error
        #error = logsumexp(delta[-1]) - correct_path_score
        self.error += error
        
        # if the error is too low, don't bother training (saves time and avoids
        # overfitting). An error of 0.01 means a log-prob of -0.01 for the right
        # tag, i.e., more than 99% probability
        # error 0.69 -> 50% probability for right tag (minimal threshold)
        # error 0.22 -> 80%
        # error 0.1  -> 90%
        if error <= 0.01:
            self.skips += 1
            return False
        
        # initialize gradients
        # dC / dftheta
        self.net_gradients = np.zeros((len(tags), self.output_size))
        # dC / dA
        #self.trans_gradients = np.zeros_like(self.transitions, float)
        self.trans_gradients = np.zeros_like(self.transitions, np.float64)
        
        # things get nasty from here
        # refer to the papers to understand what exactly is going on
        
        # compute the gradients for the last token
        # dC_logadd / ddelta_T(i) = e(delta_T(i))/Sum_k(e(delta_T(k)))
        # Compute it using the log:
        # log(e(delta_T(i))/Sum_k(e(delta_T(k)))) =
        # log(e(delta_T(i))) - log(Sum_k(e(delta_T(k)))) =
        # delta_T(i) - logsumexp(delta_T(k))
        # dC_logadd / ddelta_T(i) = e(delta_T(i) - logsumexp(delta_T(k)))
        sumlogadd = logsumexp(delta[-1])
        # negative gradients
        self.net_gradients[-1] = -np.exp(delta[-1] - sumlogadd)

        transitions_t = 0 if self.transitions is None else self.transitions[:-1].T
        
        # delta[i][j]: sum of scores of all path that assign tag j to ith-token

        # now compute the gradients for the other tokens, from last to first
        for t in range(len(scores) - 2, -1, -1):
            
            # sum the scores for all paths ending with each tag i at token t
            # with the transitions from tag i to the next tag j
            # Obtained by transposing twice
            # [delta_t-1(i)+A_j,i]T
            path_scores = (delta[t] + transitions_t).T

            # normalize over all possible tag paths using a softmax,
            # computed using log.
            # the log of the sums of exps, summed by column
            log_sum_scores = logsumexp(path_scores, 0)
            
            # softmax is the division of an exponential by the sum of all exponentials
            # (yields a probability)
            # e(delta_t-1(i)+A_i,j) / Sum_k e(delta_t-1(k)+A_k,j)
            delta_softmax = np.exp(path_scores - log_sum_scores)

            # multiply each value in the softmax by the gradient at the next tag
            # dC_logadd / ddelta_t(i) * delta_softmax
            # Attardi: negative since net_gradients[t + 1] already negative
            grad_times_softmax = self.net_gradients[t + 1] * delta_softmax
            # dC / dA_i,j
            self.trans_gradients[:-1, :] += grad_times_softmax
            
            # sum all transition gradients by row to find the network gradients
            # Sum_j(dC_logadd / ddelta_t(j) * delta_softmax)
            # Attardi: negative since grad_times_softmax already negative
            self.net_gradients[t] = np.sum(grad_times_softmax, 1)

        # find the gradients for the starting transition
        # there is only one possibility to come from, which is the sentence start
        self.trans_gradients[-1] = self.net_gradients[0]
        
        # now, add +1 to the correct path
        last_tag = self.output_size
        for token, tag in enumerate(tags):
            self.net_gradients[token][tag] += 1 # negative gradient
            if self.transitions is not None:
                self.trans_gradients[last_tag][tag] += 1 # negative gradient
            last_tag = tag
        
        return True

    @cython.boundscheck(False)
    def _calculate_gradients_wll(self, tags, scores):
        """
        Calculates the output for each token, using Word Level Likelihood.
        The aim is to minimize the word-level log-likelihood:
        C(ftheta) = logadd_j(ftheta_j) - ftheta_y,
        where y is the sequence of correct tags
        
        :returns: if True, normal gradient calculation was performed.
            If False, the error was too low and weight correction should be
            skipped.
        """
        # compute the negative gradient with respect to ftheta
        # dC / dftheta_i = e(ftheta_i)/Sum_k(e(ftheta_k))
        exponentials = np.exp(scores)
        # FIXME: use logsumexp
        # ((len(sentence), self.output_size))
        self.net_gradients = -(exponentials.T / exponentials.sum(1)).T

        # correct path and its gradient
        correct_path_score = 0
        token = 0
        for tag, net_scores in zip(tags, scores):
            self.net_gradients[token][tag] += 1 # negative gradient
            token += 1
            correct_path_score += net_scores[tag]

        # C(ftheta) = logadd_j(ftheta_j) - score(correct path)
        #error = np.log(np.sum(np.exp(scores))) - correct_path_score
        error = logsumexp(scores) - correct_path_score
        # approximate
        #error = np.max(scores) - correct_path_score
        self.error += error

        return True

    @cython.boundscheck(False)
    def _viterbi(self, np.ndarray[FLOAT_t, ndim=2] scores):
        """
        Performs a Viterbi search over the scores for each tag using
        the transitions matrix. If a matrix wasn't supplied, 
        it will return the tags with the highest scores individually.
        """
        # pretty straightforward
        if self.transitions is None or len(scores) == 1:
            return scores.argmax(1)
            
        path_scores = np.empty_like(scores)
        #path_backtrack = np.empty_like(scores, np.int64)
        path_backtrack = np.empty_like(scores, int)
        
        # now the actual Viterbi algorithm
        # first, get the scores for each tag at token 0
        # the last row of the transitions table has the scores for the first tag
        path_scores[0] = scores[0] + self.transitions[-1]
        
        output_range = np.arange(self.output_size) # outside loop. Attardi
        transitions = self.transitions[:-1]        # idem

        cdef int i
        for i in xrange(1, len(scores)):
            
            # each line contains the score until each tag t plus the transition to each other tag t'
            prev_score_and_trans = (path_scores[i - 1] + transitions.T).T
            
            # find the previous tag that yielded the max score
            path_backtrack[i] = prev_score_and_trans.argmax(0)
            path_scores[i] = prev_score_and_trans[path_backtrack[i], 
                                                  output_range] + scores[i]
            
        # now find the maximum score for the last token and follow the backtrack
        #answer = np.empty(len(scores), dtype=np.int64)
        answer = np.empty(len(scores), dtype=int)
        answer[-1] = path_scores[-1].argmax()
        self.answer_score = path_scores[-1][answer[-1]]
        previous_tag = path_backtrack[-1][answer[-1]]
        
        for i in range(len(scores) - 2, 0, -1):
            answer[i] = previous_tag
            previous_tag = path_backtrack[i][previous_tag]
        
        answer[0] = previous_tag
        return answer
    
    def set_validation_data(self, list validation_sentences=None,
                            list validation_tags=None):
        """
        Sets the data to be used during validation. If this function is not
        called before training, the training data is used to measure performance.
        
        :param validation_sentences: sentences to be used in validation.
        :param validation_tags: tags for the validation sentences.
        """
        self.validation_sentences = validation_sentences
        self.validation_tags = validation_tags 
    
    def set_learning_rate_decay(self, float decay_factor=1.0):
        """
        Sets the network to use learning rate decay. 
        
        The learning rate at each iteration t is determined as:
        initial_rate / (1 + t * decay_factor)
        
        with t starting from 0
        """
        self.use_learning_rate_decay = True
        self.decay_factor = decay_factor
    
    def decrease_learning_rates(self, epoch):
        """
        Apply the learning rate decay, if the network was configured to use it.
        """
        if not self.use_learning_rate_decay or epoch == 0:
            return
        
        # multiplying the last rate by this adjustment is equivalent to
        # initial_rate / (1 + t * decay_factor)
        # and we don't need to store the initial rates
        factor = (1.0 + (epoch - 1) * self.decay_factor) / (1 + epoch * self.decay_factor)
        self.learning_rate *= factor
        self.learning_rate_features *= factor
        if self.transitions is not None:
            self.learning_rate_trans *= factor
    
    def train(self, list sentences, list tags, 
              int epochs, int epochs_between_reports=0,
              float desired_accuracy=0):
        """
        Trains the network to tag sentences.
        
        :param sentences: a list of 2-dim numpy arrays, where each item 
            encodes a sentence. Each item in a sentence has the 
            indices to its features.
        :param tags: a list of 1-dim numpy arrays, where each item has
            the tags of the sentences.
        :param epochs: number of training epochs
        :param epochs_between_reports: number of epochs to wait between
            reports about the training performance. 0 means no reports.
        :param desired_accuracy: training stops if the desired accuracy
            is reached. Ignored if 0.
        """
        logger = logging.getLogger("Logger")
        logger.info("Training for up to %d epochs" % epochs)
        top_accuracy = 0
        last_accuracy = 0
        last_error = np.Infinity
        self.num_tokens = sum(len(sent) for sent in sentences)
        
        if self.validation_sentences is None:
            self.validation_sentences = sentences
            self.validation_tags = tags
        
        for i in range(epochs):
            self.decrease_learning_rates(i)      
            self._train_epoch(sentences, tags)
            self._validate()
            
            # normalize error
            self.error = self.error / self.num_tokens if self.num_tokens else np.Infinity
            
            # Attardi: save model
            if self.accuracy > top_accuracy:
                top_accuracy = self.accuracy
                self.save()
                logger.debug("Saved model")
            elif self.use_learning_rate_decay:
                # this iteration didn't bring improvements; load the last saved model
                # before continuing training with a lower rate
                self._load_parameters()
                        
            if (epochs_between_reports > 0 and i % epochs_between_reports == 0) \
                or self.accuracy >= desired_accuracy > 0 \
                or (self.accuracy < last_accuracy and self.error > last_error):
                
                self._print_epoch_report(i + 1)
                
                if self.accuracy >= desired_accuracy > 0\
                        or (self.error > last_error and self.accuracy < last_accuracy):
                    break
                
            last_accuracy = self.accuracy
            last_error = self.error
        
        self.num_tokens = 0
            
    def _print_epoch_report(self, int num):
        """
        Reports the status of the network in the given training
        epoch, including error and accuracy.
        """
        logger = logging.getLogger("Logger")
        logger.info("%d epochs   Error: %f   Accuracy: %f   " \
            "%d corrections skipped   " \
            "learning rate: %f" % (num,
                                   self.error,
                                   self.accuracy,
                                   self.skips,
                                   self.learning_rate))
    
    def _train_epoch(self, list sentences, list tags):
        """
        Trains for one epoch with all examples.
        """
        self.error = 0
        self.skips = 0
        self.float_errors = 0
        
        # shuffle data
        # get the random number generator state in order to shuffle
        # sentences and their tags in the same order
        random_state = np.random.get_state()
        np.random.shuffle(sentences)
        np.random.set_state(random_state)
        np.random.shuffle(tags)
        
        for sent, sent_tags in zip(sentences, tags):
            try:
                self._tag_sentence(sent, sent_tags)
            except FloatingPointError:
                # just ignore the sentence in case of an overflow
                self.float_errors += 1

    def _validate(self):
        """Perform validation on validation data and estimate accuracy"""
        hits = 0
        
        for sent, gold_tags in zip(self.validation_sentences, self.validation_tags):
            answer = self.tag_sentence(sent)
            hits += np.count_nonzero(answer == gold_tags)
        
        # self.num_tokens stores number of tokens in training sentences
        num_tokens = sum(len(sent) for sent in self.validation_sentences)
        self.accuracy = float(hits) / num_tokens

    def _backpropagate(self, sentence):
        """
        Backpropagate the gradients of the cost.
        """
        # f_1 = input_sent_values
        # f_2 = M_1 f_1 + b_2 = layer2_values
        # f_3 = hardTanh(f_2) = hidden_values
        # f_4 = M_2 f_3 + b_4

        # For l = 4..1 do:
        # dC / dtheta_l = df_l / dtheta_l dC / df_l		(19)
        # dC / df_l-1 = df_l / df_l-1 dC / df_l			(20)

        """
        Compute the gradients of the cost for each layer
        """
        # layer 4: output layer
        # dC / dW_4 = dC / df_4 f_3.T				(22)
        # (len, output_size).T (len, hidden_size) = (output_size, hidden_size)
        cdef np.ndarray[FLOAT_t, ndim=2] output_deltas
        output_deltas = self.net_gradients.T.dot(self.hidden_sent_values)

        # dC / db_4 = dC / df_4					(22)
        # (output_size) += ((len(sentence), output_size))
        # sum by column, i.e. all changes through the sentence
        output_bias_deltas = self.net_gradients.sum(0)

        # dC / df_3 = M_2.T dC / df_4				(23)
        #  (len, output_size) (output_size, hidden_size) = (len, hidden_size)
        dCdf_3 = self.net_gradients.dot(self.output_weights)

        # layer 3: HardTanh layer
        # no weights to adjust

        # dC / df_2 = hardtanhd(f_2) * dC / df_3
        # (len, hidden_size) (len, hidden_size)
        # FIXME: this goes quickly to 0.
        dCdf_2 = hardtanhd(self.layer2_sent_values) * dCdf_3

        # df_2 / df_1 = M_1

        # layer 2: linear layer
        # dC / dW_2 = dC / df_2 f_1.T				(22)
        cdef np.ndarray[FLOAT_t, ndim=2] hidden_deltas
        # (len, hidden_size).T (len, input_size) = (hidden_size, input_size)
        hidden_deltas = dCdf_2.T.dot(self.input_sent_values)

        # dC / db_2 = dC / df_2					(22)
        # sum by column contribution by each token
        hidden_bias_deltas = dCdf_2.sum(0)

        # dC / df_1 = M_1.T dC / df_2
        cdef np.ndarray[FLOAT_t, ndim=2] input_gradients
        # (len, hidden_size) (hidden_size, input_size) = (len, input_size)
        input_gradients = dCdf_2.dot(self.hidden_weights)

        """
        Adjust the weights. 
        """
        self.output_weights += output_deltas * self.learning_rate
        self.output_bias += output_bias_deltas * self.learning_rate
        self.hidden_weights += hidden_deltas * self.learning_rate
        self.hidden_bias += hidden_bias_deltas * self.learning_rate

        """
        Adjust the features indexed by the input window.
        """
        # the deltas that will be applied to the feature tables
        # they are in the same sequence as the network receives them, i.e.
        # [token1-table1][token1-table2][token2-table1][token2-table2] (...)
        # input_size = num features * window (e.g. 60 * 5). Attardi
        cdef np.ndarray[FLOAT_t, ndim=2] input_deltas
        # (len, input_size)
        input_deltas = input_gradients * self.learning_rate_features
        
        padded_sentence = np.concatenate((self.pre_padding,
                                          sentence,
                                          self.pos_padding))
        
        cdef np.ndarray[INT_t, ndim=1] features
        cdef np.ndarray[FLOAT_t, ndim=2] table
        cdef int start, end, t
        cdef int i, j

        for i, w_deltas in enumerate(input_deltas):
            # for each window (w_deltas: 300, features: 5)
            # this tracks where the deltas for the next table begins
            start = 0
            for features in padded_sentence[i:i+self.word_window_size]:
                # select the columns for each feature_tables (t: 3)
                for t, table in enumerate(self.feature_tables):
                    end = start + table.shape[1]
                    table[features[t]] += w_deltas[start:end]
                    start = end

        # Adjusts the transition scores table with the calculated gradients.
        if self.transitions is not None:
            self.transitions += self.trans_gradients * self.learning_rate_trans

    def _load_parameters(self):
        """
        Loads weights, feature tables and transition tables previously saved.
        """
        data = np.load(self.network_filename)
        self.hidden_weights = data['hidden_weights']
        self.hidden_bias = data['hidden_bias']
        self.output_weights = data['output_weights']
        self.output_bias = data['output_bias']
        self.feature_tables = list(data['feature_tables'])
        
        # check if transitions isn't None (numpy saves everything as an array)
        if data['transitions'].shape != ():
            self.transitions = data['transitions']
        else:
            self.transitions = None
    
    def save(self):
        """
        Saves the neural network to a file.
        It will save the weights, biases, sizes, padding, 
        and feature tables.
        """
        np.savez(self.network_filename, hidden_weights=self.hidden_weights,
                 output_weights=self.output_weights,
                 hidden_bias=self.hidden_bias, output_bias=self.output_bias,
                 word_window_size=self.word_window_size, 
                 input_size=self.input_size, hidden_size=self.hidden_size,
                 output_size=self.output_size, padding_left=self.padding_left,
                 padding_right=self.padding_right, transitions=self.transitions,
                 feature_tables=self.feature_tables)


    @classmethod
    def load_from_file(cls, filename):
        """
        Loads the neural network from a file.
        It will load weights, biases, sizes, padding, 
        and feature tables.
        """
        data = np.load(filename)
        #print filename
        # cython classes don't have the __dict__ attribute
        # so we can't do an elegant self.__dict__.update(data)
        hidden_weights = data['hidden_weights']
        hidden_bias = data['hidden_bias']
        output_weights = data['output_weights']
        output_bias = data['output_bias']
        
        word_window_size = data['word_window_size']
        input_size = data['input_size']
        hidden_size = data['hidden_size']
        output_size = data['output_size']
        if 'transitions' in data:
            transitions = data['transitions']
        else:
            transitions = None

        nn = Network(word_window_size, input_size, hidden_size, output_size,
                     hidden_weights, hidden_bias, output_weights, output_bias,
                     transitions)
        
        nn.padding_left = data['padding_left']
        nn.padding_right = data['padding_right']
        nn.pre_padding = np.array((nn.word_window_size // 2) * [nn.padding_left])
        nn.pos_padding = np.array((nn.word_window_size // 2) * [nn.padding_right])
        nn.feature_tables = list(data['feature_tables'])
        nn.network_filename = filename
        
        return nn 
        
# include the files for other networks
# this comes here after the Network class has already been defined
include "networkconv.pyx"
include "networkdependencyconv.pyx"



"""
string2string search
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""


"""
This module contains a wrapper for the Faiss library by Facebook AI Research.
"""

from collections import Counter  
from typing import List, Union, Optional, Dict, Any
import os
import copy
import logging
import transformers
os.environ["TOKENIZERS_PARALLELISM"] = "false"

from nltkor.make_requirement import make_requirement
try:
    import torch
    from transformers import AutoTokenizer, AutoModel, XLNetTokenizer
    import pandas as pd
    from datasets import Dataset
    # import protobuf
except ImportError:
    requirment = ['torch', 'transformers>=4.8.2', 'pandas', 'datasets', "protobuf", 'sentencepiece']
    file_path = make_requirement(requirment)
    raise Exception(f"""
    Need to install Libraries, please pip install below libraries
    \t pip install transformers>=4.8.2
    \t pip install torch
    \t pip install pandas
    \t pip install datasets
    \t pip install protobuf
    \t pip install sentencepiece
    Or, use pip install requirement.txt
    \t  pip install -r {file_path}
    """)

# from nltk.search.kobert_tokenizer import KoBERTTokenizer


class FaissSearch:
    def __new__(cls,
            mode = None,
            model_name_or_path: str = 'klue/bert-base',
            tokenizer_name_or_path: str = 'klue/bert-base',
            embedding_type: str = 'last_hidden_state',
            device: str = 'cpu'
            ) -> None:
        if mode == 'sentence':
            return FaissSearch_SenEmbed(model_name_or_path=model_name_or_path, embedding_type=embedding_type)
        elif mode == 'word':
            return FaissSearch_WordEmbed(model_name_or_path=model_name_or_path, embedding_type=embedding_type)
        elif mode == 'splade':
            return FaissSearch_Splade(model_name_or_path=model_name_or_path, embedding_type=embedding_type)
        else:
            raise ValueError("choice 'sentence' or 'word' or 'splade'")



class FaissSearch_SenEmbed:
    def __init__(self,
        model_name_or_path: str = 'klue/bert-base',
        tokenizer_name_or_path: str = 'klue/bert-base',
        embedding_type: str = 'last_hidden_state',
        device: str = 'cpu',
        ) -> None:
        """
        This function initializes the wrapper for the FAISS library, which is used to perform semantic search.


        .. attention::

            * If you use this class, please make sure to cite the following paper:

                .. code-block:: latex

                    @article{johnson2019billion,
                        title={Billion-scale similarity search with {GPUs}},
                        author={Johnson, Jeff and Douze, Matthijs and J{\'e}gou, Herv{\'e}},
                        journal={IEEE Transactions on Big Data},
                        volume={7},
                        number={3},
                        pages={535--547},
                        year={2019},
                        publisher={IEEE}
                    }

            * The code is based on the following GitHub repository:
                https://github.com/facebookresearch/faiss

        Arguments:
            model_name_or_path (str, optional): The name or path of the model to use. Defaults to 'facebook/bart-large'.
            tokenizer_name_or_path (str, optional): The name or path of the tokenizer to use. Defaults to 'facebook/bart-large'.
            device (str, optional): The device to use. Defaults to 'cpu'.

        Returns:
            None
        """

        # Set the device
        self.device = device

        # If the tokenizer is not specified, use the model name or path
        if tokenizer_name_or_path is None:
            tokenizer_name_or_path = model_name_or_path

        # Load the tokenizer
        if tokenizer_name_or_path == 'skt/kobert-base-v1':
            # self.tokenizer = KoBERTTokenizer.from_pretrained(tokenizer_name_or_path)
            self.tokenizer = XLNetTokenizer.from_pretrained(tokenizer_name_or_path)
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)

        # Load the model
        self.model = AutoModel.from_pretrained(model_name_or_path).to(self.device)

        # Set the model to evaluation mode (since we do not need the gradients)
        self.model.eval()

        # Initialize the dataset
        self.dataset = None

    
    # Auxiliary function to get the last hidden state
    def get_last_hidden_state(self,
        embeddings: torch.Tensor,
    ) -> torch.Tensor:
        """
        This function returns the last hidden state (e.g., [CLS] token's) of the input embeddings.

        Arguments:
            embeddings (torch.Tensor): The input embeddings.

        Returns:
            torch.Tensor: The last hidden state.
        """

        # Get the last hidden state
        last_hidden_state = embeddings.last_hidden_state

        # Return the last hidden state
        return last_hidden_state[:, 0, :]


    # Auxiliary function to get the mean pooling
    def get_mean_pooling(self,
        embeddings: torch.Tensor,
    ) -> torch.Tensor:
        """
        This function returns the mean pooling of the input embeddings.

        Arguments:
            embeddings (torch.Tensor): The input embeddings.

        Returns:
            torch.Tensor: The mean pooling.
        """

        # Get the mean pooling
        mean_pooling = embeddings.last_hidden_state.mean(dim=1)

        # Return the mean pooling
        return mean_pooling


    # Get the embeddings
    def get_embeddings(self,
        text: Union[str, List[str]],
        embedding_type: str = 'last_hidden_state',
        batch_size: int = 8,
        num_workers: int = 4,
    ) -> torch.Tensor:
        """
        This function returns the embeddings of the input text.

        Arguments:
            text (Union[str, List[str]]): The input text.
            embedding_type (str, optional): The type of embedding to use. Defaults to 'last_hidden_state'.
            batch_size (int, optional): The batch size to use. Defaults to 8.
            num_workers (int, optional): The number of workers to use. Defaults to 4.

        Returns:
            torch.Tensor: The embeddings.

        Raises:
            ValueError: If the embedding type is invalid.
        """

        # Check if the embedding type is valid
        if embedding_type not in ['last_hidden_state', 'mean_pooling']:
            raise ValueError(f'Invalid embedding type: {embedding_type}. Only "last_hidden_state" and "mean_pooling" are supported.')

        # Tokenize the input text
        encoded_text = self.tokenizer(
            text,
            padding=True,
            truncation=True,
            return_tensors='pt',
        )

        # Move the input text to the device
        encoded_text = encoded_text.to(self.device)

        # encoded_inputs = {k: v.to(self.device) for k, v in encoded_inputs.items()}

        # Get the embeddings
        with torch.no_grad():
            embeddings = self.model(**encoded_text)

        # Get the proper embedding type
        if embedding_type == 'last_hidden_state':
            # Get the last hidden state
            embeddings = self.get_last_hidden_state(embeddings)
        elif embedding_type == 'mean_pooling':
            # Get the mean pooling
            embeddings = self.get_mean_pooling(embeddings)

        # Return the embeddings
        return embeddings


    # Add FAISS index
    def add_faiss_index(self,
        column_name: str = 'embeddings',
        metric_type: Optional[int] = None,
        batch_size: int = 8,
        **kwargs,
    ) -> None:
        """
        This function adds a FAISS index to the dataset.

        Arguments:
            column_name (str, optional): The name of the column containing the embeddings. Defaults to 'embeddings'.
            index_type (str, optional): The index type to use. Defaults to 'Flat'.
            metric_type (str, optional): The metric type to use. Defaults to 'L2'.

        Returns:
            None

        Raises:
            ValueError: If the dataset is not initialized.
        """

        # Check if the dataset is initialized
        if self.dataset is None:
            raise ValueError('The dataset is not initialized. Please initialize the dataset first.')

        print('Adding FAISS index...')
        self.dataset.add_faiss_index(
            column_name,
            # metric_type=metric_type,
            # device=self.device,
            # batch_size=batch_size,
            faiss_verbose=True,
            # **kwargs,
        )


    def save_faiss_index(self,
        index_name: str,
        file_path: str,
    ) -> None:
        """
        This function saves the FAISS index to the specified file path.
            * This is a wrapper function for the `save_faiss_index` function in the `Dataset` class.

        Arguments:
            index_name (str): The name of the FAISS index  (e.g., "embeddings")
            file_path (str): The file path to save the FAISS index.

        Returns:
            None

        Raises:
            ValueError: If the dataset is not initialized.
        """

        # Check if the dataset is initialized
        if self.dataset is None:
            raise ValueError('The dataset is not initialized. Please initialize the dataset first.')

        print('Saving FAISS index...')
        self.dataset.save_faiss_index(index_name=index_name, file=file_path)


    def load_faiss_index(self,
        index_name: str,
        file_path: str,
        device: str = 'cpu',
    ) -> None:
        """
        This function loads the FAISS index from the specified file path.
            * This is a wrapper function for the `load_faiss_index` function in the `Dataset` class.

        Arguments:
            index_name (str): The name of the FAISS index  (e.g., "embeddings")
            file_path (str): The file path to load the FAISS index from.
            device (str, optional): The device to use ("cpu" or "cuda") (default: "cpu").

        Returns:
            None

        Raises:
            ValueError: If the dataset is not initialized.
        """

        # Check if the dataset is initialized
        if self.dataset is None:
            raise ValueError('The dataset is not initialized. Please initialize the dataset first.')

        print('Loading FAISS index...')
        self.dataset.load_faiss_index(index_name=index_name, file=file_path, device=device)


    # Initialize the corpus using a dictionary or pandas DataFrame or HuggingFace Datasets object
    def initialize_corpus(self,
        corpus: Union[Dict[str, List[str]], pd.DataFrame, Dataset],
        section: str = 'text',
        index_column_name: str = 'embeddings',
        embedding_type: str = 'last_hidden_state',
        batch_size: Optional[int] = None,
        num_workers: Optional[int] = None,
        save_path: Optional[str] = None,
    ) -> Dataset:
        """
        This function initializes a dataset using a dictionary or pandas DataFrame or HuggingFace Datasets object.

        Arguments:
            dataset_dict (Dict[str, List[str]]): The dataset dictionary.
            section (str): The section of the dataset to use whose embeddings will be used for semantic search (e.g., 'text', 'title', etc.) (default: 'text').
            index_column_name (str): The name of the column containing the embeddings (default: 'embeddings')
            embedding_type (str): The type of embedding to use (default: 'last_hidden_state').
            batch_size (int, optional): The batch size to use (default: 8).
            max_length (int, optional): The maximum length of the input sequences.
            num_workers (int, optional): The number of workers to use.
            save_path (Optional[str], optional): The path to save the dataset (default: None).

        Returns:
            Dataset: The dataset object (HuggingFace Datasets).

        Raises:
            ValueError: If the dataset is not a dictionary or pandas DataFrame or HuggingFace Datasets object.
        """

        # Create the dataset
        if isinstance(corpus, dict):
            self.dataset = Dataset.from_dict(corpus)
        elif isinstance(corpus, pd.DataFrame):
            self.dataset = Dataset.from_pandas(corpus)
        elif isinstance(corpus, Dataset):
            self.dataset = corpus
        else:
            raise ValueError('The dataset must be a dictionary or pandas DataFrame.')

        # Set the embedding_type
        self.embedding_type = embedding_type


        # Map the section of the dataset to the embeddings
        self.dataset = self.dataset.map(
            lambda x: {
                index_column_name: self.get_embeddings(x[section], embedding_type=self.embedding_type).detach().cpu().numpy()[0]
                },
            # batched=True,
            batch_size=batch_size,
            num_proc=num_workers,
        )

        # Save the dataset
        if save_path is not None:
            self.dataset.to_json(save_path)

        # Add FAISS index
        self.add_faiss_index(
            column_name=index_column_name,
        )

        # Return the dataset
        return self.dataset


    # Initialize the dataset using a JSON file
    def load_dataset_from_json(self,
        json_path: str,
    ) -> Dataset:
        """
        This function loads a dataset from a JSON file.

        Arguments:
            json_path (str): The path to the JSON file.

        Returns:
            Dataset: The dataset.
        """

        # Load the dataset
        self.dataset = Dataset.from_json(json_path)

        # Return the dataset
        return self.dataset


    # Search for the most similar elements in the dataset, given a query
    def search(self,
        query: str,
        k: int = 1,
        index_column_name: str = 'embeddings',
    ) -> pd.DataFrame:
        """
        This function searches for the most similar elements in the dataset, given a query.

        Arguments:
            query (str): The query.
            k (int, optional): The number of elements to return  (default: 1).
            index_column_name (str, optional): The name of the column containing the embeddings (default: 'embeddings')

        Returns:
            pd.DataFrame: The most similar elements in the dataset (text, score, etc.), sorted by score.

        Remarks:
            The returned elements are dictionaries containing the text and the score.
        """

        # Get the embeddings of the query
        query_embeddings = self.get_embeddings([query], embedding_type=self.embedding_type).detach().cpu().numpy()

        # Search for the most similar elements in the dataset
        scores, similar_elts = self.dataset.get_nearest_examples(
            index_name=index_column_name,
            query=query_embeddings,
            k=k,
        )

        # Convert the results to a pandas DataFrame
        results_df = pd.DataFrame.from_dict(similar_elts)

        # Add the scores
        results_df['score'] = scores

        # Sort the results by score
        results_df.sort_values("score", ascending=True, inplace=True)

        # Return the most similar elements
        return results_df



# FAISS Splade + ICT library wrapper class
class FaissSearch_Splade(FaissSearch_SenEmbed):
    def __init__(self,
        model_name_or_path: str = 'klue/bert-base',
        tokenizer_name_or_path: str = 'klue/bert-base',
        embedding_type: str = 'last_hidden_state',
        device: str = 'cpu',
        ) -> None:
        r"""
        This function initializes the wrapper for the FAISS library, which is used to perform semantic search.


        .. attention::

            * If you use this class, please make sure to cite the following paper:

                .. code-block:: latex

                    @article{johnson2019billion,
                        title={Billion-scale similarity search with {GPUs}},
                        author={Johnson, Jeff and Douze, Matthijs and J{\'e}gou, Herv{\'e}},
                        journal={IEEE Transactions on Big Data},
                        volume={7},
                        number={3},
                        pages={535--547},
                        year={2019},
                        publisher={IEEE}
                    }

            * The code is based on the following GitHub repository:
                https://github.com/facebookresearch/faiss

        Arguments:
            model_name_or_path (str, optional): The name or path of the model to use. Defaults to 'facebook/bart-large'.
            tokenizer_name_or_path (str, optional): The name or path of the tokenizer to use. Defaults to 'facebook/bart-large'.
            device (str, optional): The device to use. Defaults to 'cpu'.

        Returns:
            None
        """

        # Set the device
        self.device = device

        # If the tokenizer is not specified, use the model name or path
        if tokenizer_name_or_path is None:
            tokenizer_name_or_path = model_name_or_path

        # Load the tokenizer
        if tokenizer_name_or_path == 'skt/kobert-base-v1':
            # self.tokenizer = KoBERTTokenizer.from_pretrained(tokenizer_name_or_path)
            self.tokenizer = XLNetTokenizer.from_pretrained(tokenizer_name_or_path)
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)

        # Load the model
        self.model = transformers.BertForMaskedLM.from_pretrained(model_name_or_path).to(self.device)

        # Set the model to evaluation mode (since we do not need the gradients)
        self.model.eval()

        # Initialize the dataset
        self.dataset = None


    # Get the embeddings
    def get_embeddings(self,
        text: Union[str, List[str]],
        embedding_type: str = 'last_hidden_state',
        batch_size: int = 8,
        num_workers: int = 4,
    ) -> torch.Tensor:
        """
        This function returns the embeddings of the input text.

        Arguments:
            text (Union[str, List[str]]): The input text.
            embedding_type (str, optional): The type of embedding to use. Defaults to 'last_hidden_state'.
            batch_size (int, optional): The batch size to use. Defaults to 8.
            num_workers (int, optional): The number of workers to use. Defaults to 4.

        Returns:
            torch.Tensor: The embeddings.

        Raises:
            ValueError: If the embedding type is invalid.
        """

        # Check if the embedding type is valid
        if embedding_type not in ['last_hidden_state', 'mean_pooling']:
            raise ValueError(f'Invalid embedding type: {embedding_type}. Only "last_hidden_state" and "mean_pooling" are supported.')

        # Tokenize the input text
        encoded_text = self.tokenizer(
            text,
            padding=True,
            truncation=True,
            return_tensors='pt',
        )

        # Move the input text to the device
        encoded_text = encoded_text.to(self.device)

        # encoded_inputs = {k: v.to(self.device) for k, v in encoded_inputs.items()}

        # Get the embeddings
        with torch.no_grad():
            embeddings = self.model(**encoded_text)
        
        # Get the last hidden state
        embeddings = embeddings['logits']
        
        embeddings = torch.sum(torch.log(1+torch.relu(embeddings)) * encoded_text['attention_mask'].unsqueeze(-1), dim=1)
        e_norm = torch.nn.functional.normalize(embeddings, p=2, dim=1, eps=1e-8)

        # Return the embeddings
        return e_norm



# FAISS word embedding library wrapper class
class FaissSearch_WordEmbed(FaissSearch_SenEmbed):
    def __init__(self,
        model_name_or_path: str = 'klue/bert-base',
        tokenizer_name_or_path: str = 'klue/bert-base',
        embedding_type: str = 'last_hidden_state',
        device: str = 'cpu',
        ) -> None:
        r"""
        This function initializes the wrapper for the FAISS library, which is used to perform semantic search.


        .. attention::

            * If you use this class, please make sure to cite the following paper:

                .. code-block:: latex

                    @article{johnson2019billion,
                        title={Billion-scale similarity search with {GPUs}},
                        author={Johnson, Jeff and Douze, Matthijs and J{\'e}gou, Herv{\'e}},
                        journal={IEEE Transactions on Big Data},
                        volume={7},
                        number={3},
                        pages={535--547},
                        year={2019},
                        publisher={IEEE}
                    }

            * The code is based on the following GitHub repository:
                https://github.com/facebookresearch/faiss

        Arguments:
            model_name_or_path (str, optional): The name or path of the model to use. Defaults to 'facebook/bart-large'.
            tokenizer_name_or_path (str, optional): The name or path of the tokenizer to use. Defaults to 'facebook/bart-large'.
            device (str, optional): The device to use. Defaults to 'cpu'.

        Returns:
            None
        """

        # Set the device
        self.device = device

        # If the tokenizer is not specified, use the model name or path
        if tokenizer_name_or_path is None:
            tokenizer_name_or_path = model_name_or_path
        
        # Load the tokenizer
        if tokenizer_name_or_path == 'skt/kobert-base-v1':
            # self.tokenizer = KoBERTTokenizer.from_pretrained(tokenizer_name_or_path)
            self.tokenizer = XLNetTokenizer.from_pretrained(tokenizer_name_or_path)
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)

        # Load the model
        self.model = AutoModel.from_pretrained(model_name_or_path).to(self.device)


        # Set the model to evaluation mode (since we do not need the gradients)
        self.model.eval()

        # Initialize the dataset
        self.dataset = None


    # Get the embeddings (new code)
    def get_doc_embeddings(self,
        #text: Union[str, List[str]],
        text=None,
        embedding_type: str = 'last_hidden_state',
        batch_size: int = 8,
        num_workers: int = 4,
    ) -> torch.Tensor:
        """
        This function returns the embeddings of the input text.

        Arguments:
            text (Union[str, List[str]]): The input text.
            embedding_type (str, optional): The type of embedding to use. Defaults to 'last_hidden_state'.
            batch_size (int, optional): The batch size to use. Defaults to 8.
            num_workers (int, optional): The number of workers to use. Defaults to 4.

        Returns:
            torch.Tensor: The embeddings.

        Raises:
            ValueError: If the embedding type is invalid.
        """
        
        # Check if the embedding type is valid
        if embedding_type not in ['last_hidden_state', 'mean_pooling']:
            raise ValueError(f'Invalid embedding type: {embedding_type}. Only "last_hidden_state" and "mean_pooling" are supported.')

        ids_dict = {}
        # Tokenize the input text
        for sentence in text['text']:
            encoded_text = self.tokenizer(
                sentence,
                padding=False,
                truncation=True,
                return_tensors='pt',
                add_special_tokens=False
            )
            # Move the input text to the device
            encoded_text = encoded_text.to(self.device)
            token_ids_list = encoded_text['input_ids'].tolist()
            token_ids_list = token_ids_list[0]
            for ids in token_ids_list:
                if ids not in ids_dict.keys():
                    ids_dict[ids] = [sentence]
                else:
                    if text not in ids_dict[ids]:
                        ids_dict[ids].append(sentence)
        # Get the embeddings
        embedding_dict = {}
        self.model.eval()
        for key, value in ids_dict.items():
            embed = self.model(torch.tensor([[key]]), output_hidden_states=True).hidden_states[-1][:,0,:].detach()
            embedding_dict[embed] = value
        
        # Return the embeddings
        return embedding_dict


    # Get the embeddings (new code)
    def get_query_embeddings(self,
        text: Union[str, List[str]],
        embedding_type: str = 'last_hidden_state',
        batch_size: int = 8,
        num_workers: int = 4,
    ) -> torch.Tensor:
        """
        This function returns the embeddings of the input text.

        Arguments:
            text (Union[str, List[str]]): The input text.
            embedding_type (str, optional): The type of embedding to use. Defaults to 'last_hidden_state'.
            batch_size (int, optional): The batch size to use. Defaults to 8.
            num_workers (int, optional): The number of workers to use. Defaults to 4.

        Returns:
            torch.Tensor: The embeddings.

        Raises:
            ValueError: If the embedding type is invalid.
        """

        # Check if the embedding type is valid
        if embedding_type not in ['last_hidden_state', 'mean_pooling']:
            raise ValueError(f'Invalid embedding type: {embedding_type}. Only "last_hidden_state" and "mean_pooling" are supported.')

        # Tokenize the input text
        encoded_text = self.tokenizer(
            text,
            padding=False,
            truncation=True,
            return_tensors='pt',
            add_special_tokens=False,
        )
        
        # Move the input text to the device
        encoded_text = encoded_text.to(self.device)

        token_ids_list = encoded_text['input_ids'].tolist()
        token_ids_list = token_ids_list[0]
        tensor_list = [torch.tensor([[value]]) for value in token_ids_list]
        
        # Get the embeddings
        embeds = []
        self.model.eval()
        for index, tensor in enumerate(tensor_list):
            embed = self.model(tensor, output_hidden_states=True).hidden_states[-1][:,0,:].detach().cpu().numpy()
            embeds.append(embed)

        # Return the embeddings
        return embeds

    
    # Initialize the corpus using a dictionary or pandas DataFrame or HuggingFace Datasets object
    def initialize_corpus(self,
        corpus: Union[Dict[str, List[str]], pd.DataFrame, Dataset],
        section: str = 'text',
        index_column_name: str = 'embeddings',
        embedding_type: str = 'last_hidden_state',
        batch_size: Optional[int] = None,
        num_workers: Optional[int] = None,
        save_path: Optional[str] = None,
    ) -> Dataset:
        """
        This function initializes a dataset using a dictionary or pandas DataFrame or HuggingFace Datasets object.

        Arguments:
            dataset_dict (Dict[str, List[str]]): The dataset dictionary.
            section (str): The section of the dataset to use whose embeddings will be used for semantic search (e.g., 'text', 'title', etc.) (default: 'text').
            index_column_name (str): The name of the column containing the embeddings (default: 'embeddings')
            embedding_type (str): The type of embedding to use (default: 'last_hidden_state').
            batch_size (int, optional): The batch size to use (default: 8).
            max_length (int, optional): The maximum length of the input sequences.
            num_workers (int, optional): The number of workers to use.
            save_path (Optional[str], optional): The path to save the dataset (default: None).

        Returns:
            Dataset: The dataset object (HuggingFace Datasets).

        Raises:
            ValueError: If the dataset is not a dictionary or pandas DataFrame or HuggingFace Datasets object.
        """

        # corpus = { 'text': [...] } -> form_dict
        
        # Set the embedding_type
        self.embedding_type = embedding_type
        
        # get embedding dict
        embedding_dict = self.get_doc_embeddings(text=corpus, embedding_type=self.embedding_type)

        data = {
                'text' : embedding_dict.values(),
                'embeddings': []
                }

        for embed in embedding_dict.keys():
            embed_list = embed.tolist()
            data['embeddings'].append(embed_list[0])

        
        if isinstance(data, dict):
            self.dataset = Dataset.from_dict(data)
        elif isinstance(data, pd.DataFrame):
            self.dataset = Dataset.from_pandas(data)
        elif isinstance(data, Dataset):
            self.dataset = corpus
        else:
            raise ValueError('The dataset must be a dictionary or pandas DataFrame.')
        
        # Save the dataset
        if save_path is not None:
            self.dataset.to_json(save_path)
        
        # Add FAISS index
        self.add_faiss_index(
            column_name=index_column_name,
        )

        # Return the dataset
        return self.dataset


    # Search for the most similar elements in the dataset, given a query
    def search(self,
        query: str,
        k: int = 1,
        index_column_name: str = 'embeddings',
    ) -> pd.DataFrame:
        """
        This function searches for the most similar elements in the dataset, given a query.

        Arguments:
            query (str): The query.
            k (int, optional): The number of elements to return  (default: 1).
            index_column_name (str, optional): The name of the column containing the embeddings (default: 'embeddings')

        Returns:
            pd.DataFrame: The most similar elements in the dataset (text, score, etc.), sorted by score.

        Remarks:
            The returned elements are dictionaries containing the text and the score.
        """

        # Get the embeddings of the query
        query_embeddings = self.get_query_embeddings([query], embedding_type=self.embedding_type)

        # query_embedding이랑 self.dataset['embeddings'] 값 비교
        scores = []
        similar_elts = []
        for query in query_embeddings:
            # Search for the most similar elements in the dataset
            score, similar_elt = self.dataset.get_nearest_examples(
                index_name=index_column_name,
                query=query,
                k=k,
            )
            scores.append(score)
            similar_elts.append(similar_elt)
        

        text_list = []
        for item in similar_elts:
            for text in item['text']:
                text_list.append(text)
        
        flat_list = [sentence for sublist in text_list for sentence in sublist]
        count = Counter(flat_list)
        count = dict(count.most_common(5))
         
        sorted_dict = dict(sorted(count.items(), key=lambda x: x[1], reverse=True))
        # Convert the results to a pandas DataFrame
        results_df = pd.DataFrame({'text': sorted_dict.keys() , 'freq': sorted_dict.values()})
        
        # Return the most similar elements
        return results_df



#!/usr/env python
# -*- coding: utf-8 -*-

"""
Base class for reading NLP tagging data.
"""

import os
import sys
import re
import abc
import logging
import numpy as np
import chardet
import _pickle
from collections import Counter

from . import attributes
from . import metadata
from . import config
from .word_dictionary import WordDictionary
from .attributes import get_capitalization
from .utils import PickleConverter

class FileNotFoundException(IOError):
	"""
	Dummy class for indicating file not found instead of 
	the broad IOError.
	"""
	pass

pickle_converter = PickleConverter()

def load_tag_dict(filename):
	"""
	Load a tag dictionary from a file containing one tag
	per line.
	"""
	tag_dict = {}
	with open(filename, 'rb') as f: 
			raw_data = f.read(1024)
			detected = chardet.detect(raw_data).get('encoding', 'utf-8')
	with open(filename, 'rt', encoding = detected) as f:
		code = 0
		for tag in f:
			tag = tag.strip()
			if tag:
				tag_dict[tag] = code
				code += 1
	
	return tag_dict

def load_morph_lexicon(filename):
	if not os.path.exists(filename):
		pickle_converter.convert_morph_lexicon(filename)

	with open(filename, 'rb') as f:
		return _pickle.load(f)


def save_tag_dict(filename, tag_dict):
	"""
	Save the given tag dictionary to the given file. Dictionary
	is saved with one tag per line, in the order of their codes.
	"""
	ordered_keys = sorted(tag_dict, key=tag_dict.get)
	ordered_keys = sorted(ordered_keys)
	text = '\n'.join(ordered_keys)
	with open(filename, 'wt') as f:
		f.write(text)


class TaggerReader(object):
	"""
	Abstract class extending TextReader with useful functions
	for tagging tasks. 
	"""
	__metaclass__ = abc.ABCMeta
	
	def __init__(self, md=None, load_dictionaries=True):
		'''
		This class shouldn't be used directly. The constructor only
		provides method calls for subclasses. Subclasses should call
		this constructor after initializing the `task` attribute.
		'''
		self._set_metadata(md)
		self.codified = False
		self._converter = None
		
		if load_dictionaries:
			self.load_or_create_dictionary()	# vocabulary
			self.load_or_create_tag_dict()		# tagset
			if self.task == 'pos':
				self.load_co_lexicon()					# complicated morph lexicon
				self.load_morph_lexicon()					# user morph lexicon
				self.load_prob_dict()					# probability data
	
	@abc.abstractmethod
	def task(self):
		"""
		The task the tagger reads data for.
		Must be defined in subclasses.
		"""
		return None
	
	def load_or_create_dictionary(self):
		"""
		Try to load the vocabulary from the default location. If the vocabulary
		file is not available, create a new one from the sentences available
		and save it.
		"""
		try:
			self.load_dictionary()
		except FileNotFoundException:
			self.generate_dictionary(minimum_occurrences=1)
			#self.generate_dictionary(minimum_occurrences=2)
			self.save_dictionary()
	
	def load_or_create_tag_dict(self):
		"""
		Try to load the tag dictionary from the default location. If the dictinaty
		file is not available, scan the available sentences and create a new one. 
		"""
		key = '%s_tag_dict' % self.task
		filename = self.md.paths[key]
		if os.path.isfile(filename):
			self.load_tag_dict(filename)
			return
		
		tags = {tag for sent in self.sentences for _, tag in sent}
		self.tag_dict = {tag: code for code, tag in enumerate(tags)}
		self.save_tag_dict(filename)
	
	def load_morph_lexicon(self):
		"""
		Try to load the morph lexicon to have to be captured from the default location. 
		"""
		logger = logging.getLogger("Logger")

		key = '%s_morph_lexicon' % self.task
		filename = self.md.paths[key]
		if os.path.isfile(filename):
			print(filename)
			self.load_morph_lexicon(filename)
			return
		else:
			logger.info("Can not find % " % filename)

	def load_co_lexicon(self):
		"""
		Try to load the morph pattern lexicon from the default location. 
		"""
		logger = logging.getLogger("Logger")

		key = '%s_co_lexicon' % self.task
		filename = self.md.paths[key]
		if os.path.isfile(filename):
			self.load_co_lexicon(filename)
			return
		else:
			logger.info("Can not find % " % filename)

	def load_prob_dict(self):
		"""
		Try to load the morph pattern lexicon from the default location. 
		"""
		logger = logging.getLogger("Logger")

		key = '%s_prob_dict' % self.task
		filename = self.md.paths[key]
		if os.path.isfile(filename):
			self.load_prob_dict(filename)
			return
		else:
			logger.info("Can not find % " % filename)



	def generate_dictionary(self, dict_size=None, minimum_occurrences=1):
		"""
		Generates a token dictionary based on the given sentences.
		
		:param dict_size: Max number of tokens to be included in the dictionary.
		:param minimum_occurrences: Minimum number of times that a token must
			appear in the text in order to be included in the dictionary. 
		"""
		logger = logging.getLogger("Logger")
				
		tokens = [token for sent in self.sentences for token, _ in sent]
		self.word_dict = WordDictionary(tokens, dict_size, minimum_occurrences)
		logger.info("Created dictionary with %d types" % self.word_dict.num_tokens)
		
	def get_inverse_tag_dictionary(self):
		"""
		Returns a version of the tag dictionary that maps numbers to tags.
		Used for consulting the meaning of the network's output.
		"""
		tuples = [(x[1], x[0]) for x in self.tag_dict.items()]
		ret = dict(tuples)
		
		return ret
	
	def codify_sentence(self, sentence):
		"""
		Converts a given sentence into the indices used by the neural network.
		
		:param sentence: a sequence of tokens, already tokenized
		"""
		if self._converter is None:
			self.create_converter()
		return np.array([self.converter.convert(t) for t in sentence])
	
	def codify_sentences(self):
		logger = logging.getLogger("Logger")
		logger.info("data structuring.")
		"""
		Converts each token in each sequence into indices to their feature vectors
		in feature matrices. The previous sentences as text are not accessible anymore.
		"""
		if self._converter is None:
			self.create_converter()
		
		new_sentences = []
		self.tags = []
		rare_tag_value = self.tag_dict.get(self.rare_tag)
		
		for sent in self.sentences:
			new_sent = []
			sentence_tags = []
			
			for token, tag in sent:
				new_token = self.converter.convert(token)
				new_sent.append(new_token)
				sentence_tags.append(self.tag_dict.get(tag, rare_tag_value))
			
			new_sentences.append(np.array(new_sent))
			self.tags.append(np.array(sentence_tags))
		
		self.sentences = new_sentences
		self.codified = True
	
	def get_word_counter(self):
		"""
		Returns a Counter object with word type occurrences.
		"""
		c = Counter(token.lower() for sent in self.sentences for token, _ in sent)
		return c
	
	def get_tag_counter(self):
		"""
		Returns a Counter object with tag occurrences.
		"""
		c = Counter(tag for sent in self.sentences for _, tag in sent)
		return c
	
	def save_tag_dict(self, filename=None, tag_dict=None):
		"""
		Saves a tag dictionary to a file as a list of tags.
		
		:param tag_dict: the dictionary to save. If None, the default
			tag_dict for the class will be saved.
		:param filename: the file where the dictionary should be saved.
			If None, the class default tag_dict filename will be used.
		"""
		if tag_dict is None:
			tag_dict = self.tag_dict
		if filename is None:
			key = '%s_tag_dict' % self.task
			filename = self.md.paths[key]
		
		save_tag_dict(filename, tag_dict)
	
	def load_tag_dict(self, filename=None):
		"""
		Load the tag dictionary from the default file and assign
		it to the tag_dict attribute. 
		"""
		if filename is None:
			key = '%s_tag_dict' % self.task
			filename = self.md.paths[key]
			
		self.tag_dict = load_tag_dict(filename)

	def get_os_filename(self, filename):
		name, ext = os.path.splitext(filename)
		if os.name == "nt":
			return f"{name}-win{ext}"
		elif sys.platform == "darwin":
			return f"{name}-mac{ext}"
		elif sys.platform.startswith("linux"):
			return f"{name}-linux{ext}"
		return filename
		
	def load_morph_lexicon(self, filename=None):
		"""
		Load the morph dictionary from the default file and assign
		it to the morph_tag attribute. 
		"""
		if filename is None:
			key = '%s_morph_lexicon' % self.task
			filename = self.md.paths[key]
			
		self.morph_lexicon = load_morph_lexicon(filename)

	def load_co_lexicon(self, filename=None):
		"""
		Load the tag dictionary from the default file and assign
		it to the tag_dict attribute. 
		"""
		if filename is None:
			key = '%s_co_lexicon' % self.task
			filename = self.get_os_filename(self.md.paths[key])

		if not os.path.exists(filename):
			raise FileNotFoundError(f"{filename}")

		with open(filename, 'rb') as f:
			self.co_lexicon = _pickle.load(f)
		
	def load_prob_dict(self, filename=None):
		"""
		Load the tag dictionary from the default file and assign
		it to the tag_dict attribute. 
		"""
		if filename is None:
			key = '%s_prob_dict' % self.task
			filename = self.get_os_filename(self.md.paths[key])
		
		if not os.path.exists(filename):
			raise FileNotFoundError(f"{filename}")

		with open(filename, 'rb') as f:
			self.prob_dict = _pickle.load(f)
	
	   
	def _set_metadata(self, md):
		if md is None:
			#metadata not provided = using global data_dir for files
			self.md = metadata.Metadata(self.task, config.FILES)
		else:
			self.md = md
		
	def add_text(self, text):
		"""
		Adds more text to the reader. The text must be a sequence of sequences of 
		tokens.
		"""
		self.sentences.extend(text)
	
	def load_dictionary(self):
		"""Read a file with a word list and create a dictionary."""
		logger = logging.getLogger("Logger")
		logger.info("Loading vocabulary")
		
		# try to load vocabulary specific for the task
		key = 'vocabulary_%s' % self.task
		filename = self.md.paths[key]
		if not os.path.isfile(filename):
			# fallback to generic vocabulary
			filename = self.md.paths['vocabulary']
			if not os.path.isfile(filename):
				raise FileNotFoundException()
		
		words = []
		with open(filename, 'rb') as f: 
			raw_data = f.read(1024)
			detected = chardet.detect(raw_data).get('encoding', 'utf-8')
		with open(filename, 'rt', encoding = detected) as f:
			for word in f:
				#word = unicode(word, 'utf-8').strip()
				word = word.strip()
				if word:
					words.append(word)
		
		wd = WordDictionary.init_from_wordlist(words)
		self.word_dict = wd
		logger.info("Done. Dictionary size is %d types" % wd.num_tokens)
	
	def save_dictionary(self, filename=None):
		"""
		Saves the reader's word dictionary as a list of words.
		
		:param filename: path to the file to save the dictionary. 
			if not given, it will be saved in the default nlpnet
			data directory.
		"""
		logger = logging.getLogger("Logger")
		if filename is None:
			key = 'vocabulary_%s' % self.task
			filename = self.md.paths[key]
		
		self.word_dict.save(filename)
		logger.info("Dictionary saved in %s" % filename)
	
	def create_affix_list(self, prefix_or_suffix, max_size, min_occurrences):
		"""
		Handle the creation of suffix and prefix lists.
		
		Check if there exists an affix list in the data directory. If there isn't,
		create a new one based on the training sentences.
		
		:param prefix_or_suffix: string 'prefix' or 'suffix'
		"""
		affix_type = prefix_or_suffix.lower()
		assert affix_type == 'suffix' or affix_type == 'prefix' 
		
		filename = self.md.paths['%ses' % affix_type]
		if os.path.isfile(filename):
			return
		
		logger = logging.getLogger("Logger")
		affixes_all_lengths = []
		
		# only get the affix size n from words with length at least (n+1)
		types = {re.sub(r'\d', '9', token.lower()) 
				 for sent in self.sentences for token, _ in sent}
		
		for length in range(1, max_size + 1):
			if affix_type == 'suffix':
				c = Counter(type_[-length:]
							for type_ in types
							if len(type_) > length)
			else:
				c = Counter(type_[:length]
							for type_ in types
							if len(type_) > length)
			affixes_this_length = [affix for affix in c 
								   if c[affix] >= min_occurrences]
			affixes_all_lengths.extend(affixes_this_length)
		
		logger.info('Created a list of %d %ses.' % (len(affixes_all_lengths), affix_type))
		text = '\n'.join(affixes_all_lengths)
		with open(filename, 'wt') as f:
			f.write(text)
	
	@property
	def converter(self):
		"""
		Return the token converter, which transforms tokens into their feature
		vector indices. If it doesn't exist, one is created. 
		"""
		if self._converter is None:
			self.create_converter()
		
		return self._converter
	
	@converter.setter
	def converter(self, value):
		self._converter = value
	
	def create_converter(self):
		"""
		Sets up the token converter, which is responsible for transforming tokens into their
		feature vector indices
		"""
		def add_affix_extractors(affix):
			"""
			Helper function that works for both suffixes and prefixes.
			The parameter affix should be 'suffix' or 'prefix'.
			"""
			loader_function = getattr(attributes.Affix, 'load_%ses' % affix)
			loader_function(self.md)
			
			# deal with gaps between sizes (i.e., if there are sizes 2, 3, and 5)
			codes = getattr(attributes.Affix, '%s_codes' % affix)
			sizes = sorted(codes)

			getter = getattr(attributes.Affix, 'get_%s' % affix)
			for size in sizes:
				
				# size=size because if we don't use it, lambda sticks to the last value of 
				# the loop iterator size
				def f(word, size=size):
					return getter(re.sub(r'\d', '9', word), size)
				
				self.converter.add_extractor(f)
		
		self._converter = attributes.TokenConverter()
		self.converter.add_extractor(self.word_dict.get)
		if self.md.use_caps:
			self.converter.add_extractor(get_capitalization)
		if self.md.use_prefix:
			add_affix_extractors('prefix')
		if self.md.use_suffix:
			add_affix_extractors('suffix')



import sys
from setuptools import Extension
from setuptools import setup
from setuptools import find_packages
from setuptools import dist
import os
import subprocess
subprocess.check_call([sys.executable, "-m", "pip", "install", "Cython"])
subprocess.check_call([sys.executable, "-m", "pip", "install", "numpy>=1.23.5,<=1.26.4"])
from Cython.Build import cythonize
import numpy as np

try:
    subprocess.check_call([sys.executable, "-m", "pip", "install", "fasttext"])
    fasttext_package = "fasttext"
except subprocess.CalledProcessError:
    fasttext_package = None
    print("WARNING: `fasttext` installation failed.")
    print("Try installing `fasttext-wheel` instead: pip install fasttext-wheel")

install_requires=[
    'Cython',
    'numpy>=1.23.5,<=1.26.4',
    'regex',
    'tqdm>=4.40.0',
    'joblib',
    'requests',
    'nltk > 3.0',
    'pyarrow',
    'beautifulSoup4',
    'faiss-cpu==1.7.3',
    'datasets',
    'torch',
    'dill<0.3.9',
    'scikit-learn>=0.22.1',
    'transformers==4.42.2',
    'protobuf',
    'sentencepiece',
    'pandas',
    'bert_score',
    'chardet',
    'GPUtil'
    ]

if fasttext_package:
    install_requires.append(fasttext_package)

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

extra_files = package_files('nltkor/sejong')
extra_files = extra_files+package_files('nltkor/tag')

module1 = cythonize([
    Extension(
        "nltkor.tag.libs.network",
        ["nltkor/tag/libs/network.pyx"],
        include_dirs=['.', np.get_include()]
    )
])


setup(
  name='nltkor',
  version='1.2.16',
	url='https://modi.changwon.ac.kr/air_cwnu/nlp_tool/nltk_ko.git',
  packages=find_packages(exclude=[]),
  python_requires='>=3.7',
  install_requires=install_requires,
  package_data={'': extra_files},
	ext_modules=module1,
  include_package_data=True,

  tests_require=["pytest"],
  classifiers=[
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
      "Typing :: Typed",
  ],
  keywords=[
      "string matching",
      "pattern matching",
      "edit distance",
      "string to string correction",
      "string to string matching",
      "Levenshtein edit distance",
      "Hamming distance",
      "Damerau-Levenshtein distance",
      "Jaro-Winkler distance",
      "longest common subsequence",
      "longest common substring",
      "dynamic programming",
      "approximate string matching",
      "semantic similarity",
      "natural language processing",
      "NLP",
      "information retrieval",
      "rouge",
      "sacrebleu",
      "bertscore",
      "bartscore",
      "fasttext",
      "glove",
      "cosine similarity",
      "Smith-Waterman",
      "Needleman-Wunsch",
      "Hirschberg",
      "Karp-Rabin",
      "Knuth-Morris-Pratt",
      "Boyer-Moore",
  ],
)



# coding=utf-8
# Copyright 2021 SKT AI Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional
from nltkor.make_requirement import make_requirement
try:
    from transformers.tokenization_utils import AddedToken
    from transformers import XLNetTokenizer
    from transformers import SPIECE_UNDERLINE
    import sentencepiece
except ImportError:
    requirement = ['transformers>=4.8.2', 'sentencepiece']
    file_path = make_requirement(requirement)
    raise Exception(f"""
    Need to install Libraries, please pip install below libraries
    \t pip install transformers>=4.8.2
    \t pip install sentencepiece
    Or, use pip install requirement.txt
    \t  pip install -r {file_path}
    """)

class KoBERTTokenizer(XLNetTokenizer):
    padding_side = "right"

    def __init__(
        self,
        vocab_file,
        do_lower_case=False,
        remove_space=True,
        keep_accents=False,
        bos_token="[CLS]",
        eos_token="[SEP]",
        unk_token="[UNK]",
        sep_token="[SEP]",
        pad_token="[PAD]",
        cls_token="[CLS]",
        mask_token="[MASK]",
        additional_special_tokens=None,
        sp_model_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> None:
        # Mask token behave like a normal word, i.e. include the space before it
        mask_token = (
            AddedToken(mask_token, lstrip=True, rstrip=False)
            if isinstance(mask_token, str)
            else mask_token
        )

        self.sp_model_kwargs = {} if sp_model_kwargs is None else sp_model_kwargs

        super().__init__(
            vocab_file,
            do_lower_case=do_lower_case,
            remove_space=remove_space,
            keep_accents=keep_accents,
            bos_token=bos_token,
            eos_token=eos_token,
            unk_token=unk_token,
            sep_token=sep_token,
            pad_token=pad_token,
            cls_token=cls_token,
            mask_token=mask_token,
            additional_special_tokens=additional_special_tokens,
            sp_model_kwargs=self.sp_model_kwargs,
            **kwargs,
        )
        self._pad_token_type_id = 0

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks by concatenating and
        adding special tokens. An XLNet sequence has the following format:
        - single sequence: ``<cls> X <sep>``
        - pair of sequences: ``<cls> A <sep> B <sep>``
        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs to which the special tokens will be added.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.
        Returns:
            :obj:`List[int]`: List of `input IDs <../glossary.html#input-ids>`__ with the appropriate special tokens.
        """
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]
        if token_ids_1 is None:
            return cls + token_ids_0 + sep
        return cls + token_ids_0 + sep + token_ids_1 + sep

    def _tokenize(self, text: str) -> List[str]:
        """Tokenize a string."""
        text = self.preprocess_text(text)
        pieces = self.sp_model.encode(text, out_type=str, **self.sp_model_kwargs)
        new_pieces = []
        for piece in pieces:
            if len(piece) > 1 and piece[-1] == str(",") and piece[-2].isdigit():
                cur_pieces = self.sp_model.EncodeAsPieces(
                    piece[:-1].replace(SPIECE_UNDERLINE, "")
                )
                if (
                    piece[0] != SPIECE_UNDERLINE
                    and cur_pieces[0][0] == SPIECE_UNDERLINE
                ):
                    if len(cur_pieces[0]) == 1:
                        cur_pieces = cur_pieces[1:]
                    else:
                        cur_pieces[0] = cur_pieces[0][1:]
                cur_pieces.append(piece[-1])
                new_pieces.extend(cur_pieces)
            else:
                new_pieces.append(piece)

        return new_pieces

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks by concatenating and
        adding special tokens. An XLNet sequence has the following format:

        - single sequence: ``<cls> X <sep> ``
        - pair of sequences: ``<cls> A <sep> B <sep>``

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs to which the special tokens will be added.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.

        Returns:
            :obj:`List[int]`: List of `input IDs <../glossary.html#input-ids>`__ with the appropriate special tokens.
        """
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]
        if token_ids_1 is None:
            return cls + token_ids_0 + sep
        return cls + token_ids_0 + sep + token_ids_1 + sep

    def create_token_type_ids_from_sequences(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Create a mask from the two sequences passed to be used in a sequence-pair classification task. An XLNet
        sequence pair mask has the following format:

        ::

            0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1
            | first sequence    | second sequence |

        If :obj:`token_ids_1` is :obj:`None`, this method only returns the first portion of the mask (0s).

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.

        Returns:
            :obj:`List[int]`: List of `token type IDs <../glossary.html#token-type-ids>`_ according to the given
            sequence(s).
        """
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]
        if token_ids_1 is None:
            return len(cls + token_ids_0 + sep) * [0]
        return len(cls + token_ids_0 + sep) * [0] + len(token_ids_1 + sep) * [1]



# -*- coding: utf-8 -*- 

"""
Configuration data for the system.
"""

import os

data_dir = None
FILES = {}

def get_config_paths(directory):
    """Sets the data directory containing the data for the models."""
    assert os.path.isdir(directory), 'Invalid data directory'

    return { key: os.path.join(directory, value) for key, value in [ 
				# cross-task data
				('.', '.'), #for data_dir access
				
				# vocabulary file used as a fallback if a reader doesn't have a specific one
				('vocabulary'                  , 'vocabulary.txt'),
				('type_features'               , 'types-features.npy'),
				('termvectors'                 , 'termvectors.txt'),

				# POS
				('network_pos'                 , 'pos-network.npz'),
				('network_text_pos'            , 'pos-network.txt'),
				('pos_tags'                    , 'pos-tags.txt'),
				('pos_tag_dict'                , 'pos-tags.txt'),
				('pos_co_lexicon'              , 'pos-co-lexicon.pickle'),
				('pos_morph_lexicon'           , 'pos-morph-lexicon.pickle'),
				('pos_prob_dict'               , 'pos-prob-dict.pickle'),
				('pos_morph_lexicon_txt'       , 'pos-morph-lexicon.txt'),
				('suffix'                      , 'suffixes.txt'),
				('suffixes'                    , 'suffixes.txt'),
				('prefix'                      , 'prefixes.txt'),
				('prefixes'                    , 'prefixes.txt'),
				('metadata_pos'                , 'metadata-pos.pickle'),
				('metadata_text_pos'           , 'metadata-pos.txt'),
				('type_features_pos'           , 'types-features-pos.npy'),
				('caps_features_pos'           , 'caps-features-pos.npy'),
				('suffix_features_pos'         , 'suffix-features-pos.npy'),
				('prefix_features_pos'         , 'prefix-features-pos.npy'),
				('vocabulary_pos'              , 'vocabulary-pos.txt'),
				
				# NER
				('network_ner'                 , 'ner-network.npz'),
				('network_text_ner'            , 'ner-network.txt'),
				('ner_tags'                    , 'ner-tags.txt'),
				('ner_tag_dict'                , 'ner-tags.txt'),
				('ner_morph_lexicon'           , 'ner-morph-lexicon.txt'),
				('suffix'                      , 'suffixes.txt'),
				('suffixes'                    , 'suffixes.txt'),
				('prefix'                      , 'prefixes.txt'),
				('prefixes'                    , 'prefixes.txt'),
				('metadata_ner'                , 'metadata-ner.pickle'),
				('metadata_text_ner'           , 'metadata-ner.txt'),
				('type_features_ner'           , 'types-features-ner.npy'),
				('caps_features_ner'           , 'caps-features-ner.npy'),
				('suffix_features_ner'         , 'suffix-features-ner.npy'),
				('prefix_features_ner'         , 'prefix-features-ner.npy'),
				('vocabulary_ner'              , 'vocabulary-ner.txt'),

				# WSD
				('network_wsd'                 , 'wsd-network.npz'),
				('network_text_wsd'            , 'wsd-network.txt'),
				('wsd_tags'                    , 'wsd-tags.txt'),
				('wsd_tag_dict'                , 'wsd-tags.txt'),
				('wsd_morph_lexicon'           , 'wsd-morph-lexicon.txt'),
#				('suffix'                      , 'suffixes.txt'),
#				('suffixes'                    , 'suffixes.txt'),
#				('prefix'                      , 'prefixes.txt'),
#				('prefixes'                    , 'prefixes.txt'),
				('metadata_wsd'                , 'metadata-wsd.pickle'),
				('metadata_text_wsd'           , 'metadata-wsd.txt'),
				('type_features_wsd'           , 'types-features-wsd.npy'),
				('caps_features_wsd'           , 'caps-features-wsd.npy'),
				('suffix_features_wsd'         , 'suffix-features-wsd.npy'),
				('prefix_features_wsd'         , 'prefix-features-wsd.npy'),
				('vocabulary_wsd'              , 'vocabulary-wsd.txt'),

				# dependency
				('network_labeled_dependency', 'ldep-network.npz'),
				('network_text_labeled_dependency', 'ldep-network.txt'),
				('type_features_labeled_dependency', 'types-features-ldep.npy'),
				('caps_features_labeled_dependency', 'caps-features-ldep.npy'),
				('pos_features_labeled_dependency', 'pos-features-ldep.npy'),
				('metadata_labeled_dependency', 'metadata-ldep.pickle'),
				('metadata_text_labeled_dependency', 'metadata-ldep.txt'),
				('dependency_tag_dict', 'dependency-tags.txt'),
				('labeled_dependency_tag_dict', 'dependency-tags.txt'),
				('vocabulary_labeled_dependency', 'vocabulary-ldep.txt'),
				
				('dependency_pos_tags', 'dep-pos-tags.txt'),
				
				('network_unlabeled_dependency', 'udep-network.npz'),
				('network_text_unlabeled_dependency', 'udep-network.txt'),
				('type_features_unlabeled_dependency', 'types-features-udep.npy'),
				('caps_features_unlabeled_dependency', 'caps-features-udep.npy'),
				('pos_features_unlabeled_dependency', 'pos-features-udep.npy'),
				('metadata_unlabeled_dependency', 'metadata-udep.pickle'),
				('metadata_text_unlabeled_dependency', 'metadata-udep.txt'),
				('vocabulary_unlabeled_dependency', 'vocabulary-udep.txt'),
				
				# chunk
				#('chunk_tag_dict'              , 'chunk-tag-dict.pickle'),
				#('chunk_tags'                  , 'chunk-tags.txt'),
				
				# SRL
				('network_srl'                  , 'srl-network.npz'),
				('network_text_srl'             , 'srl-network.txt'),
				#('network_srl_boundary'        , 'srl-id-network.npz'),
				#('network_srl_classify'        , 'srl-class-network.npz'),
				#('network_srl_predicates'      , 'srl-class-predicates.npz'),
				#('srl_iob_tag_dict'            , 'srl-tags.txt'),
				#('srl_iob_tags'                , 'srl-tags.txt'),
				('srl_tags'                    , 'srl-tags.txt'),
				#('srl_classify_tag_dict'       , 'srl-tags.txt'),
				#('srl_classify_tags'           , 'srl-tags.txt'),
				#('srl_predicates_tag_dict'     , 'srl-predicates-tags.txt'),
				#('srl_predicates_tags'         , 'srl-predicates-tags.txt'),
				('type_features_srl'            , 'types-features-srl.npy'),
				('caps_features_srl'            , 'caps-features-srl.npy'),
				('pos_features_srl'             , 'pos-features-srl.npy'),
				#('chunk_features_classify'     , 'chunk-features-class.npy'),
				#('type_features_boundary'      , 'types-features-id.npy'),
				#('caps_features_boundary'      , 'caps-features-id.npy'),
				#('pos_features_boundary'       , 'pos-features-id.npy'),
				#('chunk_features_boundary'     , 'chunk-features-id.npy'),
				#('type_features_classify'      , 'types-features-class.npy'),
				#('caps_features_classify'      , 'caps-features-class.npy'),
				#('pos_features_classify'       , 'pos-features-class.npy'),
				#('chunk_features_classify'     , 'chunk-features-class.npy'),
				#('type_features_1step'         , 'types-features-1step.npy'),
				#('caps_features_1step'         , 'caps-features-1step.npy'),
				#('pos_features_1step'          , 'pos-features-1step.npy'),
				#('chunk_features_1step'        , 'chunk-features-1step.npy'),
				#('type_features_srl_predicates', 'types-features-preds.npy'),
				#('caps_features_srl_predicates', 'caps-features-preds.npy'),
				#('pos_features_srl_predicates' , 'pos-features-preds.npy'),
				('metadata_srl'                , 'metadata-srl.pickle'),
				('metadata_text_srl'            , 'metadata-srl.txt'),
				#('metadata_srl_boundary'       , 'metadata-srl-boundary.pickle'),
				#('metadata_srl_classify'       , 'metadata-srl-classify.pickle'),
				#('metadata_srl_predicates'     , 'metadata-srl-predicates.pickle'),
				('vocabulary_srl', 'vocabulary-srl.txt'),
				#('vocabulary_srl_boundary', 'vocabulary-srl-boundary.txt'),
				#('vocabulary_srl_classify', 'vocabulary-srl-classify.txt'),
				#('vocabulary_srl_predicates', 'vocabulary-srl-predicates.txt')
				]
    }


def set_data_dir(directory):
    """Sets the global data directory containing the data for the models."""
    global data_dir, FILES
    data_dir = directory
    FILES = get_config_paths(directory)

